#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 21
void NES_mapper21::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  regs[0] = 0;
  regs[1] = 1;
  regs[2] = 2;
  regs[3] = 3;
  regs[4] = 4;
  regs[5] = 5;
  regs[6] = 6;
  regs[7] = 7;
  regs[8] = 0;

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper21::MemoryWrite(uint32 addr, uint8 data)
{
  // regs[0] ... 1K VROM bank at PPU $0000
  // regs[1] ... 1K VROM bank at PPU $0400
  // regs[2] ... 1K VROM bank at PPU $0800
  // regs[3] ... 1K VROM bank at PPU $0C00
  // regs[4] ... 1K VROM bank at PPU $1000
  // regs[5] ... 1K VROM bank at PPU $1400
  // regs[6] ... 1K VROM bank at PPU $1800
  // regs[7] ... 1K VROM bank at PPU $1C00
  // regs[8] ... $8000 Switching Mode

  switch (addr & 0xF0CF)
  {
    case 0x8000:
      {
        if(regs[8] & 0x02)
        {
          set_CPU_bank6(data);
        }
        else
        {
          set_CPU_bank4(data);
        }
      }
      break;

    case 0xA000:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0x9000:
      {
        data &= 0x03;
        if(data == 0)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 1)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 2)
        {
          set_mirroring(0,0,0,0);
        }
        else
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0x9002:
    case 0x9080:
      {
        regs[8] = data;
      }
      break;

    case 0xB000:
      {
        regs[0] = (regs[0] & 0xF0) | (data & 0x0F);
        set_PPU_bank0(regs[0]);
      }
      break;

    case 0xB002:
    case 0xB040:
      {
        regs[0] = (regs[0] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank0(regs[0]);
      }
      break;

    case 0xB001:
    case 0xB004:
    case 0xB080:
      {
        regs[1] = (regs[1] & 0xF0) | (data & 0x0F);
        set_PPU_bank1(regs[1]);
      }
      break;

    case 0xB003:
    case 0xB006:
    case 0xB0C0:
      {
        regs[1] = (regs[1] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank1(regs[1]);
      }
      break;

    case 0xC000:
      {
        regs[2] = (regs[2] & 0xF0) | (data & 0x0F);
        set_PPU_bank2(regs[2]);
      }
      break;

    case 0xC002:
    case 0xC040:
      {
        regs[2] = (regs[2] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank2(regs[2]);
      }
      break;

    case 0xC001:
    case 0xC004:
    case 0xC080:
      {
        regs[3] = (regs[3] & 0xF0) | (data & 0x0F);
        set_PPU_bank3(regs[3]);
      }
      break;

    case 0xC003:
    case 0xC006:
    case 0xC0C0:
      {
        regs[3] = (regs[3] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank3(regs[3]);
      }
      break;

    case 0xD000:
      {
        regs[4] = (regs[4] & 0xF0) | (data & 0x0F);
        set_PPU_bank4(regs[4]);
      }
      break;

    case 0xD002:
    case 0xD040:
      {
        regs[4] = (regs[4] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank4(regs[4]);
      }
      break;

    case 0xD001:
    case 0xD004:
    case 0xD080:
      {
        regs[5] = (regs[5] & 0xF0) | (data & 0x0F);
        set_PPU_bank5(regs[5]);
      }
      break;

    case 0xD003:
    case 0xD006:
    case 0xD0C0:
      {
        regs[5] = (regs[5] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank5(regs[5]);
      }
      break;

    case 0xE000:
      {
        regs[6] = (regs[6] & 0xF0) | (data & 0x0F);
        set_PPU_bank6(regs[6]);
      }
      break;

    case 0xE002:
    case 0xE040:
      {
        regs[6] = (regs[6] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank6(regs[6]);
      }
      break;

    case 0xE001:
    case 0xE004:
    case 0xE080:
      {
        regs[7] = (regs[7] & 0xF0) | (data & 0x0F);
        set_PPU_bank7(regs[7]);
      }
      break;

    case 0xE003:
    case 0xE006:
    case 0xE0C0:
      {
        regs[7] = (regs[7] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank7(regs[7]);
      }
      break;

    case 0xF000:
      {
        irq_latch = (irq_latch & 0xF0) | (data & 0x0F);
      }
      break;

    case 0xF002:
    case 0xF040:
      {
        irq_latch = (irq_latch & 0x0F) | ((data & 0x0F) << 4);
      }
      break;

    case 0xF003:
    case 0xF0C0:
      {
        irq_enabled = (irq_enabled & 0x01) * 3;
      }
      break;

    case 0xF004:
    case 0xF080:
      {
        irq_enabled = data & 0x03;
        if(irq_enabled & 0x02)
        {
          irq_counter = irq_latch;
        }
      }
      break;
  }
}

void NES_mapper21::HSync(uint32 scanline)
{
  if(irq_enabled & 0x02)
  {
    if(irq_counter == 0xFF)
    {
      irq_counter = irq_latch;
      irq_enabled = (irq_enabled & 0x01) * 3;
      parent_NES->cpu->DoIRQ();
    }
    else
    {
      irq_counter++;
    }
  }
}

#define MAP21_VROM(ptr) (((ptr)-parent_NES->ROM->get_VROM_banks()) >> 10)

void NES_mapper21::SNSS_fixup()
{
  regs[0] = MAP21_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
  regs[1] = MAP21_VROM(parent_NES->ppu->PPU_VRAM_banks[1]);
  regs[2] = MAP21_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
  regs[3] = MAP21_VROM(parent_NES->ppu->PPU_VRAM_banks[3]);
  regs[4] = MAP21_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
  regs[5] = MAP21_VROM(parent_NES->ppu->PPU_VRAM_banks[5]);
  regs[6] = MAP21_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
  regs[7] = MAP21_VROM(parent_NES->ppu->PPU_VRAM_banks[7]);
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 22
void NES_mapper22::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper22::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x9000:
      {
        data &= 0x03;
        if(data == 0)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 1)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 2)
        {
          set_mirroring(1,1,1,1);
        }
        else
        {
          set_mirroring(0,0,0,0);
        }
      }
      break;

    case 0xA000:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0xB000:
      {
        set_PPU_bank0(data >> 1);
      }
      break;

    case 0xB001:
      {
        set_PPU_bank1(data >> 1);
      }
      break;

    case 0xC000:
      {
        set_PPU_bank2(data >> 1);
      }
      break;

    case 0xC001:
      {
        set_PPU_bank3(data >> 1);
      }
      break;

    case 0xD000:
      {
        set_PPU_bank4(data >> 1);
      }
      break;

    case 0xD001:
      {
        set_PPU_bank5(data >> 1);
      }
      break;

    case 0xE000:
      {
        set_PPU_bank6(data >> 1);
      }
      break;

    case 0xE001:
      {
        set_PPU_bank7(data >> 1);
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 23
void NES_mapper23::Reset()
{
  patch = 0xFFFF;

  // for Akumajou Special - Boku Doracula Kun
  if(sum_16k_rom == 0x137ba4)
  {
    patch = 0xF00C;
  }

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);

  regs[0] = 0;
  regs[1] = 1;
  regs[2] = 2;
  regs[3] = 3;
  regs[4] = 4;
  regs[5] = 5;
  regs[6] = 6;
  regs[7] = 7;
  regs[8] = 0;

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper23::MemoryWrite(uint32 addr, uint8 data)
{
  // regs[0] ... 1K VROM bank at PPU $0000
  // regs[1] ... 1K VROM bank at PPU $0400
  // regs[2] ... 1K VROM bank at PPU $0800
  // regs[3] ... 1K VROM bank at PPU $0C00
  // regs[4] ... 1K VROM bank at PPU $1000
  // regs[5] ... 1K VROM bank at PPU $1400
  // regs[6] ... 1K VROM bank at PPU $1800
  // regs[7] ... 1K VROM bank at PPU $1C00
  // regs[8] ... $9008 swap

  switch(addr & patch)
  {
    case 0x8000:
    case 0x8004:
    case 0x8008:
    case 0x800C:
      {
        if(regs[8])
        {
          set_CPU_bank6(data);
        }
        else
        {
          set_CPU_bank4(data);
        }
      }
      break;

    case 0x9000:
      {
        if(data != 0xFF)
        {
          data &= 0x03;
          if(data == 0)
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
          else if(data == 1)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else if(data == 2)
          {
            set_mirroring(0,0,0,0);
          }
          else
          {
            set_mirroring(1,1,1,1);
          }
        }
      }
      break;

    case 0x9008:
      {
        regs[8] = data & 0x02;
      }
      break;

    case 0xA000:
    case 0xA004:
    case 0xA008:
    case 0xA00C:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0xB000:
      {
        regs[0] = (regs[0] & 0xF0) | (data & 0x0F);
        set_PPU_bank0(regs[0]);
      }
      break;

    case 0xB001:
    case 0xB004:
      {
        regs[0] = (regs[0] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank0(regs[0]);
      }
      break;

    case 0xB002:
    case 0xB008:
      {
        regs[1] = (regs[1] & 0xF0) | (data & 0x0F);
        set_PPU_bank1(regs[1]);
      }
      break;

    case 0xB003:
    case 0xB00C:
      {
        regs[1] = (regs[1] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank1(regs[1]);
      }
      break;

    case 0xC000:
      {
        regs[2] = (regs[2] & 0xF0) | (data & 0x0F);
        set_PPU_bank2(regs[2]);
      }
      break;

    case 0xC001:
    case 0xC004:
      {
        regs[2] = (regs[2] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank2(regs[2]);
      }
      break;

    case 0xC002:
    case 0xC008:
      {
        regs[3] = (regs[3] & 0xF0) | (data & 0x0F);
        set_PPU_bank3(regs[3]);
      }
      break;

    case 0xC003:
    case 0xC00C:
      {
        regs[3] = (regs[3] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank3(regs[3]);
      }
      break;

    case 0xD000:
      {
        regs[4] = (regs[4] & 0xF0) | (data & 0x0F);
        set_PPU_bank4(regs[4]);
      }
      break;

    case 0xD001:
    case 0xD004:
      {
        regs[4] = (regs[4] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank4(regs[4]);
      }
      break;

    case 0xD002:
    case 0xD008:
      {
        regs[5] = (regs[5] & 0xF0) | (data & 0x0F);
        set_PPU_bank5(regs[5]);
      }
      break;

    case 0xD003:
    case 0xD00C:
      {
        regs[5] = (regs[5] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank5(regs[5]);
      }
      break;

    case 0xE000:
      {
        regs[6] = (regs[6] & 0xF0) | (data & 0x0F);
        set_PPU_bank6(regs[6]);
      }
      break;

    case 0xE001:
    case 0xE004:
      {
        regs[6] = (regs[6] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank6(regs[6]);
      }
      break;

    case 0xE002:
    case 0xE008:
      {
        regs[7] = (regs[7] & 0xF0) | (data & 0x0F);
        set_PPU_bank7(regs[7]);
      }
      break;

    case 0xE003:
    case 0xE00C:
      {
        regs[7] = (regs[7] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank7(regs[7]);
      }
      break;

    case 0xF000:
      {
        irq_latch = (irq_latch & 0xF0) | (data & 0x0F);
      }
      break;

    case 0xF004:
      {
        irq_latch = (irq_latch & 0x0F) | ((data & 0x0F) << 4);
      }
      break;

    case 0xF008:
      {
        irq_enabled = data & 0x03;
        if(irq_enabled & 0x02)
        {
          irq_counter = irq_latch;
        }
      }
      break;

    case 0xF00C:
      {
        irq_enabled = (irq_enabled & 0x01) * 3;
      }
      break;
  }
}

void NES_mapper23::HSync(uint32 scanline)
{
  if(irq_enabled & 0x02)
  {
    if(irq_counter == 0xFF)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = irq_latch;
      irq_enabled = (irq_enabled & 0x01) * 3;
    }
    else
    {
      irq_counter++;
    }
  }
}

#define MAP23_VROM(ptr) (((ptr)-parent_NES->ROM->get_VROM_banks()) >> 10)

void NES_mapper23::SNSS_fixup()
{
  regs[0] = MAP23_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
  regs[1] = MAP23_VROM(parent_NES->ppu->PPU_VRAM_banks[1]);
  regs[2] = MAP23_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
  regs[3] = MAP23_VROM(parent_NES->ppu->PPU_VRAM_banks[3]);
  regs[4] = MAP23_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
  regs[5] = MAP23_VROM(parent_NES->ppu->PPU_VRAM_banks[5]);
  regs[6] = MAP23_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
  regs[7] = MAP23_VROM(parent_NES->ppu->PPU_VRAM_banks[7]);
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 24
void NES_mapper24::Reset()
{
  // Init ExSound
  parent_NES->apu->SelectExSound(1);

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper24::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF003)
  {
    case 0x8000:
      {
        set_CPU_bank4(data*2+0);
        set_CPU_bank5(data*2+1);
      }
      break;

    case 0xB003:
      {
        data = data & 0x0C;
        if(data == 0x00)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 0x04)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 0x08)
        {
          set_mirroring(0,0,0,0);
        }
        else if(data == 0x0C)
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0xC000:
      {
        set_CPU_bank6(data);
      }
      break;

    case 0xD000:
      {
        set_PPU_bank0(data);
      }
      break;

    case 0xD001:
      {
        set_PPU_bank1(data);
      }
      break;

    case 0xD002:
      {
        set_PPU_bank2(data);
      }
      break;

    case 0xD003:
      {
        set_PPU_bank3(data);
      }
      break;

    case 0xE000:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0xE001:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0xE002:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0xE003:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0xF000:
      {
        irq_latch = data;
      }
      break;

    case 0xF001:
      {
        irq_enabled = data & 0x03;
        if(irq_enabled & 0x02)
        {
          irq_counter = irq_latch;
        }
      }
      break;

    case 0xF002:
      {
        irq_enabled = (irq_enabled & 0x01) * 3;
      }
      break;
  }
  parent_NES->apu->ExWrite(addr, data);
}

void NES_mapper24::HSync(uint32 scanline)
{
  if(irq_enabled & 0x02)
  {
    if(irq_counter == 0xFF)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = irq_latch;
      //irq_enabled = (irq_enabled & 0x01) * 3;
    }
    else
    {
      irq_counter++;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 25
void NES_mapper25::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  for(int i = 0; i < 8; i++)
    regs[i] = 0;

  regs[8] = 0;
  regs[9] = num_8k_ROM_banks-2;
  regs[10] = 0;

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper25::MemoryWrite(uint32 addr, uint8 data)
{
  // regs[0] ... 1K VROM bank at PPU $0000
  // regs[1] ... 1K VROM bank at PPU $0400
  // regs[2] ... 1K VROM bank at PPU $0800
  // regs[3] ... 1K VROM bank at PPU $0C00
  // regs[4] ... 1K VROM bank at PPU $1000
  // regs[5] ... 1K VROM bank at PPU $1400
  // regs[6] ... 1K VROM bank at PPU $1800
  // regs[7] ... 1K VROM bank at PPU $1C00
  // regs[8] ... 8K ROM bank at CPU $8000
  // regs[9] ... 8K ROM bank at CPU $C000
  // regs[10] .. ROM Swap flag

  switch(addr & 0xF000)
  {
    case 0x8000:
      {
        if(regs[10] & 0x02)
        {
          regs[9] = data;
          set_CPU_bank6(data);
        }
        else
        {
          regs[8] = data;
          set_CPU_bank4(data);
        }
      }
      break;

    case 0xA000:
      {
        set_CPU_bank5(data);
      }
      break;
  }

  switch(addr & 0xF00F)
  {
    case 0x9000:
      {
        data &= 0x03;
        if(data == 0)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 1)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 2)
        {
          set_mirroring(0,0,0,0);
        }
        else
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0x9001:
    case 0x9004:
      {
        if((regs[10] & 0x02) != (data & 0x02))
        {
          uint8 swap = regs[8];
          regs[8] = regs[9];
          regs[9] = swap;
          set_CPU_bank4(regs[8]);
          set_CPU_bank6(regs[9]);
        }
        regs[10] = data;
      }
      break;

    case 0xB000:
      {
        regs[0] = (regs[0] & 0xF0) | (data & 0x0F);
        set_PPU_bank0(regs[0]);
      }
      break;

    case 0xB001:
    case 0xB004:
      {
        regs[1] = (regs[1] & 0xF0) | (data & 0x0F);
        set_PPU_bank1(regs[1]);
      }
      break;

    case 0xB002:
    case 0xB008:
      {
        regs[0] = (regs[0] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank0(regs[0]);
      }
      break;

    case 0xB003:
    case 0xB00C:
      {
        regs[1] = (regs[1] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank1(regs[1]);
      }
      break;

    case 0xC000:
      {
        regs[2] = (regs[2] & 0xF0) | (data & 0x0F);
        set_PPU_bank2(regs[2]);
      }
      break;

    case 0xC001:
    case 0xC004:
      {
        regs[3] = (regs[3] & 0xF0) | (data & 0x0F);
        set_PPU_bank3(regs[3]);
      }
      break;

    case 0xC002:
    case 0xC008:
      {
        regs[2] = (regs[2] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank2(regs[2]);
      }
      break;

    case 0xC003:
    case 0xC00C:
      {
        regs[3] = (regs[3] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank3(regs[3]);
      }
      break;

    case 0xD000:
      {
        regs[4] = (regs[4] & 0xF0) | (data & 0x0F);
        set_PPU_bank4(regs[4]);
      }
      break;

    case 0xD001:
    case 0xD004:
      {
        regs[5] = (regs[5] & 0xF0) | (data & 0x0F);
        set_PPU_bank5(regs[5]);
      }
      break;

    case 0xD002:
    case 0xD008:
      {
        regs[4] = (regs[4] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank4(regs[4]);
      }
      break;

    case 0xD003:
    case 0xD00C:
      {
        regs[5] = (regs[5] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank5(regs[5]);
      }
      break;

    case 0xE000:
      {
        regs[6] = (regs[6] & 0xF0) | (data & 0x0F);
        set_PPU_bank6(regs[6]);
      }
      break;

    case 0xE001:
    case 0xE004:
      {
        regs[7] = (regs[7] & 0xF0) | (data & 0x0F);
        set_PPU_bank7(regs[7]);
      }
      break;

    case 0xE002:
    case 0xE008:
      {
        regs[6] = (regs[6] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank6(regs[6]);
      }
      break;

    case 0xE003:
    case 0xE00C:
      {
        regs[7] = (regs[7] & 0x0F) | ((data & 0x0F) << 4);
        set_PPU_bank7(regs[7]);
      }
      break;

    case 0xF000:
      {
        irq_latch = (irq_latch & 0xF0) | (data & 0x0F);
      }
      break;

    case 0xF001:
    case 0xF004:
      {
        irq_enabled = data & 0x03;
        if(irq_enabled & 0x02)
        {
          irq_counter = irq_latch;
        }
      }
      break;

    case 0xF002:
    case 0xF008:
      {
        irq_latch = (irq_latch & 0x0F) | ((data & 0x0F) << 4);
      }
      break;

    case 0xF003:
    case 0xF00C:
      {
        irq_enabled = (irq_enabled & 0x01) * 3;
      }
      break;
  }
}

void NES_mapper25::HSync(uint32 scanline)
{
  if(irq_enabled & 0x02)
  {
    if(irq_counter == 0xFF)
    {
      irq_counter = irq_latch;
      parent_NES->cpu->DoIRQ();
    }
    else
    {
      irq_counter++;
    }
  }
}

#define MAP25_ROM(ptr)  (((ptr)-parent_NES->ROM->get_ROM_banks())  >> 13)
#define MAP25_VROM(ptr) (((ptr)-parent_NES->ROM->get_VROM_banks()) >> 10)

void NES_mapper25::SNSS_fixup()
{
  NES_6502::Context context;
  parent_NES->cpu->GetContext(&context);
  regs[0] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
  regs[1] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[1]);
  regs[2] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
  regs[3] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[3]);
  regs[4] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
  regs[5] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[5]);
  regs[6] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
  regs[7] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[7]);
  regs[8] = MAP25_ROM(context.mem_page[4]);
  regs[9] = MAP25_ROM(context.mem_page[6]);
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 26
void NES_mapper26::Reset()
{
  // Init ExSound
  parent_NES->apu->SelectExSound(1);

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_enabled = 0;
  irq_counter = 0;
}

void NES_mapper26::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        set_CPU_bank4(data*2+0);
        set_CPU_bank5(data*2+1);
      }
      break;

    case 0xB003:
      {
        data = data & 0x7F;
        if(data == 0x08 || data == 0x2C)
        {
          set_mirroring(1,1,1,1);
        }
        else if(data == 0x20)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 0x24)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 0x28)
        {
          set_mirroring(0,0,0,0);
        }
      }
      break;

    case 0xC000:
      {
        set_CPU_bank6(data);
      }
      break;

    case 0xD000:
      {
        set_PPU_bank0(data);
      }
      break;

    case 0xD001:
      {
        set_PPU_bank2(data);
      }
      break;

    case 0xD002:
      {
        set_PPU_bank1(data);
      }
      break;

    case 0xD003:
      {
        set_PPU_bank3(data);
      }
      break;

    case 0xE000:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0xE001:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0xE002:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0xE003:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0xF000:
      {
        irq_latch = data;
      }
      break;

    case 0xF001:
      {
        irq_enabled = (irq_enabled & 0x01) * 3;
      }
      break;

    case 0xF002:
      {
        irq_enabled = data & 0x03;
        if(irq_enabled & 0x02)
        {
          irq_counter = irq_latch;
        }
      }
      break;
  }
  
  addr = addr & 0xfffc | ((addr & 1) << 1) | ((addr & 2) >> 1);
  parent_NES->apu->ExWrite(addr, data);
}

void NES_mapper26::HSync(uint32 scanline)
{
  if(irq_enabled & 0x02)
  {
    if(irq_counter >= 0xFF)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = irq_latch;
      irq_enabled = 0;
    }
    else
    {
      irq_counter++;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 73
void NES_mapper73::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  irq_counter = 0;
  irq_enabled = 0;
}

void NES_mapper73::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        irq_counter = (irq_counter & 0xFFF0) | (data & 0x0F);
      }
      break;

    case 0x9000:
      {
        irq_counter = (irq_counter & 0xFF0F) | ((data & 0x0F) << 4);
      }
      break;

    case 0xA000:
      {
        irq_counter = (irq_counter & 0xF0FF) | ((data & 0x0F) << 8);
      }
      break;

    case 0xB000:
      {
        irq_counter = (irq_counter & 0x0FFF) | ((data & 0x0F) << 12);
      }
      break;

    case 0xC000:
      {
        irq_enabled = data;
      }
      break;

    case 0xF000:
      {
        set_CPU_bank4(data*2+0);
        set_CPU_bank5(data*2+1);
      }
      break;
  }
}

void NES_mapper73::HSync(uint32 scanline)
{
  if(irq_enabled & 0x02)
  {
    if(irq_counter > 0xFFFF - 114)
    {
      parent_NES->cpu->DoIRQ();
      irq_enabled = 0;
    }
    else
    {
      irq_counter += 114;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 85

void NES_mapper85::Reset()
{
  patch = 0;

  // Tiny Toon Adventure 2
  if(sum_16k_rom == 0x17c3b6)
  {
    patch = 1;
  }

  // Init ExSound
  parent_NES->apu->SelectExSound(2);

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(patch)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
  else
  {
    parent_NES->ppu->PPU_VRAM_banks[0] = chr_ram + 0 * 0x400;
    parent_NES->ppu->PPU_VRAM_banks[1] = chr_ram + 0 * 0x400;
    parent_NES->ppu->PPU_VRAM_banks[2] = chr_ram + 0 * 0x400;
    parent_NES->ppu->PPU_VRAM_banks[3] = chr_ram + 0 * 0x400;
    parent_NES->ppu->PPU_VRAM_banks[4] = chr_ram + 0 * 0x400;
    parent_NES->ppu->PPU_VRAM_banks[5] = chr_ram + 0 * 0x400;
    parent_NES->ppu->PPU_VRAM_banks[6] = chr_ram + 0 * 0x400;
    parent_NES->ppu->PPU_VRAM_banks[7] = chr_ram + 0 * 0x400;
  }

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper85::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF038)
  {
    case 0x8000:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x8008:
    case 0x8010:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0x9000:
      {
        set_CPU_bank6(data);
      }
      break;

    case 0x9010:
    case 0x9030:
      {
        parent_NES->apu->ExWrite(addr, data);
      }
      break;

    case 0xA000:
      {
        if(patch)
        {
          set_PPU_bank0(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[0] = chr_ram + data * 0x400;
        }
      }
      break;

    case 0xA008:
    case 0xA010:
      {
        if(patch)
        {
          set_PPU_bank1(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[1] = chr_ram + data * 0x400;
        }
      }
      break;

    case 0xB000:
      {
        if(patch)
        {
          set_PPU_bank2(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[2] = chr_ram + data * 0x400;
        }
      }
      break;

    case 0xB008:
    case 0xB010:
      {
        if(patch)
        {
          set_PPU_bank3(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[3] = chr_ram + data * 0x400;
        }
      }
      break;

    case 0xC000:
      {
        if(patch)
        {
          set_PPU_bank4(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[4] = chr_ram + data * 0x400;
        }
      }
      break;

    case 0xC008:
    case 0xC010:
      {
        if(patch)
        {
          set_PPU_bank5(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[5] = chr_ram + data * 0x400;
        }
      }
      break;

    case 0xD000:
      {
        if(patch)
        {
          set_PPU_bank6(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[6] = chr_ram + data * 0x400;
        }
      }
      break;

    case 0xD008:
    case 0xD010:
      {
        if(patch)
        {
          set_PPU_bank7(data);
        }
        else
        {
          parent_NES->ppu->PPU_VRAM_banks[7] = chr_ram + data * 0x400;
        }
      }
      break;

    case 0xE000:
      {
        data &= 0x03;
        if(data == 0x00)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 0x02)
        {
          set_mirroring(0,0,0,0);
        }
        else
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0xE008:
    case 0xE010:
      {
        irq_latch = data;
      }
      break;

    case 0xF000:
      {
        irq_enabled = data & 0x03;
        if(irq_enabled & 0x02)
        {
          irq_counter = irq_latch;
        }
      }
      break;

    case 0xF008:
    case 0xF010:
      {
        irq_enabled = (irq_enabled & 0x01) * 3;
      }
      break;
  }
}

void NES_mapper85::HSync(uint32 scanline)
{
  if(irq_enabled & 0x02)
  {
    if(irq_counter == 0xFF)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = irq_latch;
    }
    else
    {
      irq_counter++;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 87

void NES_mapper87::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper87::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(addr == 0x6000)
  {
    uint8 chr_bank = (data & 0x02) >> 1;
    
    set_PPU_bank0(chr_bank*8+0);
    set_PPU_bank1(chr_bank*8+1);
    set_PPU_bank2(chr_bank*8+2);
    set_PPU_bank3(chr_bank*8+3);
    set_PPU_bank4(chr_bank*8+4);
    set_PPU_bank5(chr_bank*8+5);
    set_PPU_bank6(chr_bank*8+6);
    set_PPU_bank7(chr_bank*8+7);
  }
}
/////////////////////////////////////////////////////////////////////

#endif
