#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 5
void NES_mapper5::Reset()
{
  wram_size = 1;

  if(sum_16k_rom == 0x1f34b0 || // Bandit Kings of Ancient China
     sum_16k_rom == 0x21489f || // Dai Koukai Jidai
     sum_16k_rom == 0x23a287 || // Ishin no Arashi
     sum_16k_rom == 0x223cea || // Nobunaga no Yabou - Sengoku Gunyuu Den
     sum_16k_rom == 0x1f2e7d || // L'Empereur (J)
     sum_16k_rom == 0x1e18ab || // L'Empereur (U)
     sum_16k_rom == 0x20be7a || // Nobunaga's Ambition 2
     sum_16k_rom == 0x1f7383 || // Suikoden - Tenmei no Chikai
     sum_16k_rom == 0x21265b)   // Uncharted Waters
  {
    wram_size = 2;
  }

  if(sum_16k_rom == 0x0e6fa4 || // Aoki Ookami to Shiroki Mejika - Genchou Hishi
     sum_16k_rom == 0x1d73da || // Nobunaga no Yabou - Bushou Fuuun Roku
                                // Romance of the 3 Kingdoms 2
     sum_16k_rom == 0x202445)   // Sangokushi 2
  {
    wram_size = 3;
  }

  // set SaveRAM
  for(uint32 i = 0; i < 0x10000; i++)
  {
    wram[i] = parent_NES->ReadSaveRAM(i);
  }
  MMC5_set_WRAM_bank(3,0);

  // Init ExSound
  parent_NES->apu->SelectExSound(8);

  // set CPU bank pointers
  set_CPU_bank4(num_8k_ROM_banks-1);
  set_CPU_bank5(num_8k_ROM_banks-1);
  set_CPU_bank6(num_8k_ROM_banks-1);
  set_CPU_bank7(num_8k_ROM_banks-1);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);

  for(i = 0; i < 8; i++)
  {
    chr_reg[i][0] = i;
    chr_reg[i][1] = (i & 0x03) + 4;
  }
  wb[3] = 0;
  wb[4] = wb[5] = wb[6] = 8;

  prg_size = 3;
  wram_protect0 = 0x02;
  wram_protect1 = 0x01;
  chr_size = 3;
  gfx_mode = 0;

  irq_enabled = 0;
  irq_status = 0;
  irq_line = 0;
}

uint8 NES_mapper5::MemoryReadLow(uint32 addr)
{
  uint8 ret = (uint8)(addr >> 8);
  if(addr == 0x5204)
  {
    ret = irq_status;
    irq_status = 0;
  }
  else if(addr == 0x5205)
  {
    ret = (uint8)((value0*value1) & 0x00FF);
  }
  else if(addr == 0x5206)
  {
    ret = (uint8)(((value0*value1) & 0xFF00) >> 8);
  }
  else if(addr >= 0x5C00 && addr <=0x5FFF)
  {
    if(gfx_mode == 2 || gfx_mode == 3)
    {
      uint8* nametable2 = parent_NES->ppu->get_namt() + 0x800;
      ret = nametable2[addr & 0x3FF];
    }
  }
  return ret;
}

void NES_mapper5::MemoryWriteLow(uint32 addr, uint8 data)
{
  uint32 i;

  switch(addr)
  {
    case 0x5100:
      {
        prg_size = data & 0x03;
      }
      break;

    case 0x5101:
      {
        chr_size = data & 0x03;
      }
      break;

    case 0x5102:
      {
        wram_protect0 = data & 0x03;
      }
      break;

    case 0x5103:
      {
        wram_protect1 = data & 0x03;
      }
      break;

    case 0x5104:
      {
        gfx_mode = data & 0x03;
      }
      break;

    case 0x5105:
      {
        for(i = 0; i < 4; i++)
        {
          parent_NES->ppu->set_PPU_VRAM(8+i, data & 0x03);
          data >>= 2;
        }
      }
      break;

    case 0x5106:
      {
        uint8* nametable3 = parent_NES->ppu->get_namt() + 0xC00;
        for(i = 0; i < 0x3C0; i++)
        {
          nametable3[i] = data;
        }
      }
      break;

    case 0x5107:
      {
        uint8* nametable3 = parent_NES->ppu->get_namt() + 0xC00;
        data &= 0x03;
        data = data | (data<<2) | (data<<4) | (data<<6);
        for(i = 0x3C0; i < 0x400; i++)
        {
          nametable3[i] = data;
        }
      }
      break;

    case 0x5113:
      {
        MMC5_set_WRAM_bank(3, data & 0x07);
      }
      break;

    case 0x5114:
    case 0x5115:
    case 0x5116:
    case 0x5117:
      {
        MMC5_set_CPU_bank(addr & 0x07, data);
      }
      break;

    case 0x5120:
    case 0x5121:
    case 0x5122:
    case 0x5123:
    case 0x5124:
    case 0x5125:
    case 0x5126:
    case 0x5127:
      {
        chr_reg[addr & 0x07][0] = data;
        sync_Chr_banks(0);
      }
      break;

    case 0x5128:
    case 0x5129:
    case 0x512A:
    case 0x512B:
      {
        chr_reg[(addr & 0x03) + 0][1] = data;
        chr_reg[(addr & 0x03) + 4][1] = data;
      }
      break;

    case 0x5200:
      {
        split_control = data;
      }
      break;

    case 0x5201:
      {
        // split scroll
      }
      break;

    case 0x5202:
      {
        split_bank = data & 0x3F;
      }
      break;

    case 0x5203:
      {
        irq_line = data;
      }
      break;

    case 0x5204:
      {
        irq_enabled = data;
      }
      break;

    case 0x5205:
      {
        value0 = data;
      }
      break;

    case 0x5206:
      {
        value1 = data;
      }
      break;

    default:
      {
        if(addr >= 0x5000 && addr <= 0x5015)
        {
          parent_NES->apu->ExWrite(addr, data);
        }
        else if(addr >= 0x5C00 && addr <= 0x5FFF)
        {
          if(gfx_mode != 3)
          {
            uint8* nametable2 = parent_NES->ppu->get_namt() + 0x800;
            nametable2[addr & 0x3FF] = data;
          }
        }
      }
      break;
  }
}

void NES_mapper5::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(wram_protect0 == 0x02 && wram_protect1 == 0x01)
  {
    if(wb[3] != 8)
    {
      wram[wb[3]*0x2000+(addr&0x1FFF)] = data;
      parent_NES->WriteSaveRAM(wb[3]*0x2000+(addr&0x1FFF), data);
    }
  }
  parent_NES->WriteSaveRAM(addr&0x1FFF, wram[addr&0x1FFF]);
}

void NES_mapper5::MemoryWrite(uint32 addr, uint8 data)
{
  if(wram_protect0 == 0x02 && wram_protect1 == 0x01)
  {
    if(addr >= 0x8000 && addr <= 0x9FFF)
    {
      if(wb[4] != 8)
      {
        wram[wb[4]*0x2000+(addr&0x1FFF)] = data;
        parent_NES->WriteSaveRAM(wb[4]*0x2000+(addr&0x1FFF), data);
      }
    }
    else if(addr >= 0xA000 && addr <= 0xBFFF)
    {
      if(wb[5] != 8)
      {
        wram[wb[5]*0x2000+(addr&0x1FFF)] = data;
        parent_NES->WriteSaveRAM(wb[5]*0x2000+(addr&0x1FFF), data);
      }
    }
    else if(addr >= 0xC000 && addr <= 0xDFFF)
    {
      if(wb[6] != 8)
      {
        wram[wb[6]*0x2000+(addr&0x1FFF)] = data;
        parent_NES->WriteSaveRAM(wb[6]*0x2000+(addr&0x1FFF), data);
      }
    }
  }
}

void NES_mapper5::HSync(uint32 scanline)
{
  if(scanline == 0)
  {
    if(parent_NES->ppu->spr_enabled() && parent_NES->ppu->bg_enabled())
    {
      irq_counter = 0;
    }
  }
  if(scanline <= 245)
  {
    if(irq_counter == irq_line)
    {
      irq_status |= 0x80;
    }
    if(parent_NES->ppu->spr_enabled() && parent_NES->ppu->bg_enabled())
    {
      irq_counter++;
    }
    if((irq_status & 0x80) && (irq_enabled & 0x80))
    {
      parent_NES->cpu->DoIRQ();
    }
  }
  else
  {
    //irq_counter = 0;
    irq_status |= 0x40;
  }
}

void NES_mapper5::MMC5_set_CPU_bank(uint8 page, uint8 bank)
{
  if(bank & 0x80)
  {
    if(prg_size == 0)
    {
      if(page == 7)
      {
        set_CPU_bank4((bank & 0x7C)+0);
        set_CPU_bank5((bank & 0x7C)+1);
        set_CPU_bank6((bank & 0x7C)+2);
        set_CPU_bank7((bank & 0x7C)+3);
        wb[4] = wb[5] = wb[6] = 8;
      }
    }
    if(prg_size == 1)
    {
      if(page == 5)
      {
        set_CPU_bank4((bank & 0x7E)+0);
        set_CPU_bank5((bank & 0x7E)+1);
        wb[4] = wb[5] = 8;
      }
      if(page == 7)
      {
        set_CPU_bank6((bank & 0x7E)+0);
        set_CPU_bank7((bank & 0x7E)+1);
        wb[6] = 8;
      }
    }
    if(prg_size == 2)
    {
      if(page == 5)
      {
        set_CPU_bank4((bank & 0x7E)+0);
        set_CPU_bank5((bank & 0x7E)+1);
        wb[4] = wb[5] = 8;
      }
      if(page == 6)
      {
        set_CPU_bank6(bank & 0x7F);
        wb[6] = 8;
      }
      if(page == 7)
      {
        set_CPU_bank7(bank & 0x7F);
      }
    }
    if(prg_size == 3)
    {
      if(page == 4)
      {
        set_CPU_bank4(bank & 0x7F);
        wb[4] = 8;
      }
      if(page == 5)
      {
        set_CPU_bank5(bank & 0x7F);
        wb[5] = 8;
      }
      if(page == 6)
      {
        set_CPU_bank6(bank & 0x7F);
        wb[6] = 8;
      }
      if(page == 7)
      {
        set_CPU_bank7(bank & 0x7F);
      }
    }
  }
  else
  {
    if(prg_size == 1)
    {
      if(page == 5)
      {
        MMC5_set_WRAM_bank(4, (bank & 0x06)+0);
        MMC5_set_WRAM_bank(5, (bank & 0x06)+1);
      }
    }
    if(prg_size == 2)
    {
      if(page == 5)
      {
        MMC5_set_WRAM_bank(4, (bank & 0x06)+0);
        MMC5_set_WRAM_bank(5, (bank & 0x06)+1);
      }
      if(page == 6)
      {
        MMC5_set_WRAM_bank(6, bank & 0x07);
      }
    }
    if(prg_size == 3)
    {
      if(page == 4)
      {
        MMC5_set_WRAM_bank(4, bank & 0x07);
      }
      if(page == 5)
      {
        MMC5_set_WRAM_bank(5, bank & 0x07);
      }
      if(page == 6)
      {
        MMC5_set_WRAM_bank(6, bank & 0x07);
      }
    }
  }
}

void NES_mapper5::MMC5_set_WRAM_bank(uint8 page, uint8 bank)
{
  if(bank != 8)
  {
    if(wram_size == 1) bank = (bank > 3) ? 8 : 0;
    if(wram_size == 2) bank = (bank > 3) ? 1 : 0;
    if(wram_size == 3) bank = (bank > 3) ? 8 : bank;
    if(wram_size == 4) bank = (bank > 3) ? 4 : bank;
  }
  wb[page] = bank;

  if(bank != 8)
  {
    NES_6502::Context context;
    parent_NES->cpu->GetContext(&context);
    context.mem_page[page] = wram + bank*0x2000;
    parent_NES->cpu->SetContext(&context);
  }
}

uint8 NES_mapper5::PPU_Latch_RenderScreen(uint8 mode, uint32 addr)
{
  uint8 ex_pal = 0;

  if(gfx_mode == 1 && mode == 1)
  {
    // ex gfx mode
    uint8* nametable2 = parent_NES->ppu->get_namt() + 0x800;
    uint32 bank = (nametable2[addr] & 0x3F) << 2;
    set_PPU_banks(bank,bank+1,bank+2,bank+3,bank,bank+1,bank+2,bank+3);
    ex_pal = ((nametable2[addr] & 0xC0) >> 4) | 0x01;
  }
  else
  {
    // normal
    sync_Chr_banks(mode);
  }
  return ex_pal;
}

void NES_mapper5::sync_Chr_banks(uint8 mode)
{
  if(chr_size == 0)
  {
    set_PPU_bank0(chr_reg[7][mode]*8+0);
    set_PPU_bank1(chr_reg[7][mode]*8+1);
    set_PPU_bank2(chr_reg[7][mode]*8+2);
    set_PPU_bank3(chr_reg[7][mode]*8+3);
    set_PPU_bank4(chr_reg[7][mode]*8+4);
    set_PPU_bank5(chr_reg[7][mode]*8+5);
    set_PPU_bank6(chr_reg[7][mode]*8+6);
    set_PPU_bank7(chr_reg[7][mode]*8+7);
  }
  else if(chr_size == 1)
  {
    set_PPU_bank0(chr_reg[3][mode]*4+0);
    set_PPU_bank1(chr_reg[3][mode]*4+1);
    set_PPU_bank2(chr_reg[3][mode]*4+2);
    set_PPU_bank3(chr_reg[3][mode]*4+3);
    set_PPU_bank4(chr_reg[7][mode]*4+0);
    set_PPU_bank5(chr_reg[7][mode]*4+1);
    set_PPU_bank6(chr_reg[7][mode]*4+2);
    set_PPU_bank7(chr_reg[7][mode]*4+3);
  }
  else if(chr_size == 2)
  {
    set_PPU_bank0(chr_reg[1][mode]*2+0);
    set_PPU_bank1(chr_reg[1][mode]*2+1);
    set_PPU_bank2(chr_reg[3][mode]*2+0);
    set_PPU_bank3(chr_reg[3][mode]*2+1);
    set_PPU_bank4(chr_reg[5][mode]*2+0);
    set_PPU_bank5(chr_reg[5][mode]*2+1);
    set_PPU_bank6(chr_reg[7][mode]*2+0);
    set_PPU_bank7(chr_reg[7][mode]*2+1);
  }
  else
  {
    set_PPU_bank0(chr_reg[0][mode]);
    set_PPU_bank1(chr_reg[1][mode]);
    set_PPU_bank2(chr_reg[2][mode]);
    set_PPU_bank3(chr_reg[3][mode]);
    set_PPU_bank4(chr_reg[4][mode]);
    set_PPU_bank5(chr_reg[5][mode]);
    set_PPU_bank6(chr_reg[6][mode]);
    set_PPU_bank7(chr_reg[7][mode]);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 10
void NES_mapper10::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  for(int i = 0; i < sizeof(regs)/sizeof(regs[0]); i++)
    regs[i] = 0;

  regs[2] = 4;

  latch_0000 = 0xFE;
  latch_1000 = 0xFE;

  set_VROM_0000();
  set_VROM_1000();
}

void NES_mapper10::PPU_Latch_FDFE(uint32 addr)
{
  if(addr & 0x1000)
  {
    latch_1000 = (addr & 0x0FF0) >> 4;
    set_VROM_1000();
  }
  else
  {
    latch_0000 = (addr & 0x0FF0) >> 4;
    set_VROM_0000();
  }
}

void NES_mapper10::set_VROM_0000()
{
  int bank_num = (latch_0000 == 0xFD) ? regs[1] : regs[2];

  bank_num <<= 2;

  set_PPU_bank0(bank_num+0);
  set_PPU_bank1(bank_num+1);
  set_PPU_bank2(bank_num+2);
  set_PPU_bank3(bank_num+3);
}

void NES_mapper10::set_VROM_1000()
{
  int bank_num = (latch_1000 == 0xFD) ? regs[3] : regs[4];

  bank_num <<= 2;

  set_PPU_bank4(bank_num+0);
  set_PPU_bank5(bank_num+1);
  set_PPU_bank6(bank_num+2);
  set_PPU_bank7(bank_num+3);
}

void NES_mapper10::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF000)
  {
    case 0xA000:
      {
        regs[0] = data;

        // 16K ROM bank at $8000
        uint8 bank_num = regs[0];
        set_CPU_bank4(bank_num*2);
        set_CPU_bank5(bank_num*2+1);
      }
      break;

    case 0xB000:
      {
        // B000-BFFF: select 4k VROM for (0000) $FD latch
        regs[1] = data;
        set_VROM_0000();
      }
      break;

    case 0xC000:
      {
        // C000-CFFF: select 4k VROM for (0000) $FE latch
        regs[2] = data;
        set_VROM_0000();
      }
      break;

    case 0xD000:
      {
        // D000-DFFF: select 4k VROM for (1000) $FD latch
        regs[3] = data;
        set_VROM_1000();
      }
      break;

    case 0xE000:
      {
        // E000-EFFF: select 4k VROM for (1000) $FE latch
        regs[4] = data;
        set_VROM_1000();
      }
      break;

    case 0xF000:
      {
        regs[5] = data;

        if(regs[5] & 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;
  }
}

void NES_mapper10::SNSS_fixup()
{
  set_VROM_0000();
  set_VROM_1000();
}

/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 47
void NES_mapper47::Reset()
{
  // clear registers FIRST!!!
  for(int i = 0; i < 8; i++) regs[i] = 0x00;

  // set CPU bank pointers
  rom_bank = 0;
  prg0 = 0;
  prg1 = 1;
  MMC3_set_CPU_banks();

  // set VROM banks
  if(num_1k_VROM_banks)
  {
    chr01 = 0;
    chr23 = 2;
    chr4  = 4;
    chr5  = 5;
    chr6  = 6;
    chr7  = 7;
    MMC3_set_PPU_banks();
  }
  else
  {
    chr01 = chr23 = chr4 = chr5 = chr6 = chr7 = 0;
  }

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper47::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(addr == 0x6000)
  {
    rom_bank = data & 0x01;
    MMC3_set_CPU_banks();
    MMC3_set_PPU_banks();
  }
}

void NES_mapper47::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xE001)
  {
    case 0x8000:
      {
        regs[0] = data;
        MMC3_set_PPU_banks();
        MMC3_set_CPU_banks();
      }
      break;

    case 0x8001:
      {
        uint32 bank_num;
        regs[1] = data;
        bank_num = regs[1];

        switch(regs[0] & 0x07)
        {
          case 0x00:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr01 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x01:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr23 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x02:
            {
              if(num_1k_VROM_banks)
              {
                chr4 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x03:
            {
              if(num_1k_VROM_banks)
              {
                chr5 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x04:
            {
              if(num_1k_VROM_banks)
              {
                chr6 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x05:
            {
              if(num_1k_VROM_banks)
              {
                chr7 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x06:
            {
              prg0 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;

          case 0x07:
            {
              prg1 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;
        }
      }
      break;

    case 0xA000:
      {
        regs[2] = data;

        if(parent_NES->ROM->get_mirroring() != NES_PPU::MIRROR_FOUR_SCREEN)
        {
          if(data & 0x01)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
        }
      }
      break;

    case 0xA001:
      {
        regs[3] = data;
      }
      break;

    case 0xC000:
      regs[4] = data;
      irq_counter = regs[4];
      break;

    case 0xC001:
      regs[5] = data;
      irq_latch = regs[5];
      break;

    case 0xE000:
      regs[6] = data;
      irq_enabled = 0;
      break;

    case 0xE001:
      regs[7] = data;
      irq_enabled = 1;
      break;
  }
}

void NES_mapper47::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(!(--irq_counter))
        {
          irq_counter = irq_latch;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}

void NES_mapper47::MMC3_set_CPU_banks()
{
  if(prg_swap())
  {
    set_CPU_bank4(rom_bank*16+14);
    set_CPU_bank5(rom_bank*16+prg1);
    set_CPU_bank6(rom_bank*16+prg0);
    set_CPU_bank7(rom_bank*16+15);
  }
  else
  {
    set_CPU_bank4(rom_bank*16+prg0);
    set_CPU_bank5(rom_bank*16+prg1);
    set_CPU_bank6(rom_bank*16+14);
    set_CPU_bank7(rom_bank*16+15);
  }
}

void NES_mapper47::MMC3_set_PPU_banks()
{
  if(num_1k_VROM_banks)
  {
    if(chr_swap())
    {
      set_PPU_bank0(rom_bank*128+chr4);
      set_PPU_bank1(rom_bank*128+chr5);
      set_PPU_bank2(rom_bank*128+chr6);
      set_PPU_bank3(rom_bank*128+chr7);
      set_PPU_bank4(rom_bank*128+chr01+0);
      set_PPU_bank5(rom_bank*128+chr01+1);
      set_PPU_bank6(rom_bank*128+chr23+0);
      set_PPU_bank7(rom_bank*128+chr23+1);
    }
    else
    {
      set_PPU_bank0(rom_bank*128+chr01+0);
      set_PPU_bank1(rom_bank*128+chr01+1);
      set_PPU_bank2(rom_bank*128+chr23+0);
      set_PPU_bank3(rom_bank*128+chr23+1);
      set_PPU_bank4(rom_bank*128+chr4);
      set_PPU_bank5(rom_bank*128+chr5);
      set_PPU_bank6(rom_bank*128+chr6);
      set_PPU_bank7(rom_bank*128+chr7);
    }
  }
}
/////////////////////////////////////////////////////////////////////

#endif
