#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 15
void NES_mapper15::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);
}

void NES_mapper15::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        if(data & 0x80)
        {
          set_CPU_bank4((data & 0x3F)*2+1);
          set_CPU_bank5((data & 0x3F)*2+0);
          set_CPU_bank6((data & 0x3F)*2+3);
          set_CPU_bank7((data & 0x3F)*2+2);
        }
        else
        {
          set_CPU_bank4((data & 0x3F)*2);
          set_CPU_bank5((data & 0x3F)*2+1);
          set_CPU_bank6((data & 0x3F)*2+2);
          set_CPU_bank7((data & 0x3F)*2+3);
        }
        if(data & 0x40)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;

    case 0x8001:
      {
        if(data & 0x80)
        {
          set_CPU_bank6((data & 0x3F)*2+1);
          set_CPU_bank7((data & 0x3F)*2);
        }
        else
        {
          set_CPU_bank6((data & 0x3F)*2);
          set_CPU_bank7((data & 0x3F)*2+1);
        }
      }
      break;

    case 0x8002:
      {
        if(data & 0x80)
        {
          set_CPU_bank4((data & 0x3F)*2+1);
          set_CPU_bank5((data & 0x3F)*2+1);
          set_CPU_bank6((data & 0x3F)*2+1);
          set_CPU_bank7((data & 0x3F)*2+1);
        }
        else
        {
          set_CPU_bank4((data & 0x3F)*2);
          set_CPU_bank5((data & 0x3F)*2);
          set_CPU_bank6((data & 0x3F)*2);
          set_CPU_bank7((data & 0x3F)*2);
        }
      }
      break;

    case 0x8003:
      {
        if(data & 0x80)
        {
          set_CPU_bank6((data & 0x3F)*2+1);
          set_CPU_bank7((data & 0x3F)*2);
        }
        else
        {
          set_CPU_bank6((data & 0x3F)*2);
          set_CPU_bank7((data & 0x3F)*2+1);
        }
        if(data & 0x40)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 41
void NES_mapper41::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper41::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(addr < 0x6800)
  {
    data = (uint8)(addr & 0xFF);

    set_CPU_bank4((data & 0x07)*4);
    set_CPU_bank5((data & 0x07)*4+1);
    set_CPU_bank6((data & 0x07)*4+2);
    set_CPU_bank7((data & 0x07)*4+3);
    regs[0] = (data & 0x04);
    regs[1] = (data & 0x18) >> 1;
    if(data & 0x20)
    {
      set_mirroring(NES_PPU::MIRROR_HORIZ);
    }
    else
    {
      set_mirroring(NES_PPU::MIRROR_VERT);
    }
  }
}

void NES_mapper41::MemoryWrite(uint32 addr, uint8 data)
{
  if(regs[0])
  {
    uint8 chr_bank = regs[1] | (data & 0x0003);
    set_PPU_bank0(chr_bank*8);
    set_PPU_bank1(chr_bank*8+1);
    set_PPU_bank2(chr_bank*8+2);
    set_PPU_bank3(chr_bank*8+3);
    set_PPU_bank4(chr_bank*8+4);
    set_PPU_bank5(chr_bank*8+5);
    set_PPU_bank6(chr_bank*8+6);
    set_PPU_bank7(chr_bank*8+7);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 44
void NES_mapper44::Reset()
{
  // clear registers FIRST!!!
  for(int i = 0; i < 8; i++) regs[i] = 0x00;

  // set game select register
  rom_bank = 0;

  // set CPU bank pointers
  prg0 = 0;
  prg1 = 1;
  MMC3_set_CPU_banks();

  // set VROM banks
  if(num_1k_VROM_banks)
  {
    chr01 = 0;
    chr23 = 2;
    chr4  = 4;
    chr5  = 5;
    chr6  = 6;
    chr7  = 7;
    MMC3_set_PPU_banks();
  }
  else
  {
    chr01 = chr23 = chr4 = chr5 = chr6 = chr7 = 0;
  }

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper44::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xE001)
  {
    case 0x8000:
      {
        regs[0] = data;
        MMC3_set_PPU_banks();
        MMC3_set_CPU_banks();
      }
      break;

    case 0x8001:
      {
        uint32 bank_num;
        regs[1] = data;
        bank_num = regs[1];

        switch(regs[0] & 0x07)
        {
          case 0x00:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr01 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x01:
            {
              if(num_1k_VROM_banks)
              {
                bank_num &= 0xfe;
                chr23 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x02:
            {
              if(num_1k_VROM_banks)
              {
                chr4 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x03:
            {
              if(num_1k_VROM_banks)
              {
                chr5 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x04:
            {
              if(num_1k_VROM_banks)
              {
                chr6 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x05:
            {
              if(num_1k_VROM_banks)
              {
                chr7 = bank_num;
                MMC3_set_PPU_banks();
              }
            }
            break;

          case 0x06:
            {
              prg0 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;

          case 0x07:
            {
              prg1 = bank_num;
              MMC3_set_CPU_banks();
            }
            break;
        }
      }
      break;

    case 0xA000:
      {
        regs[2] = data;

        if(parent_NES->ROM->get_mirroring() != NES_PPU::MIRROR_FOUR_SCREEN)
        {
          if(data & 0x01)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
        }
      }
      break;

    case 0xA001:
      {
        rom_bank = data & 0x07;
        if(rom_bank == 7)
        {
          rom_bank = 6;
        }
        MMC3_set_CPU_banks();
        MMC3_set_PPU_banks();
      }
      break;

    case 0xC000:
      regs[4] = data;
      irq_counter = regs[4];
      break;

    case 0xC001:
      regs[5] = data;
      irq_latch = regs[5];
      break;

    case 0xE000:
      regs[6] = data;
      irq_enabled = 0;
      break;

    case 0xE001:
      regs[7] = data;
      irq_enabled = 1;
      break;
  }
}

void NES_mapper44::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(!(irq_counter--))
        {
          irq_counter = irq_latch;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 45
void NES_mapper45::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  prg0 = 0;
  prg1 = 1;
  prg2 = num_8k_ROM_banks-2;
  prg3 = num_8k_ROM_banks-1;
  p[0] = prg0; p[1] = prg1; p[2] = prg2; p[3] = prg3;

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);
  chr0 = 0; chr1 = 1; chr2 = 2; chr3 = 3;
  chr4 = 4; chr5 = 5; chr6 = 6; chr7 = 7;
  c[0] = chr0; c[1] = chr1; c[2] = chr2; c[3] = chr3;
  c[4] = chr4; c[5] = chr5; c[6] = chr6; c[7] = chr7;

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;

  regs[0] = regs[1] = regs[2] = regs[3] = 0;
  regs[4] = regs[5] = regs[6] = 0;
}

void NES_mapper45::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(addr == 0x6000)
  {
    regs[regs[5]] = data;
    regs[5]= (regs[5]+1) & 0x03;
    MAP45_set_CPU_bank4(prg0);
    MAP45_set_CPU_bank5(prg1);
    MAP45_set_CPU_bank6(prg2);
    MAP45_set_CPU_bank7(prg3);
    MAP45_set_PPU_banks();
  }
}

void NES_mapper45::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xE001)
  {
    case 0x8000:
      {
        uint32 swap;
        if((data & 0x40) != (regs[6] & 0x40))
        {
          swap = prg0; prg0 = prg2; prg2 = swap;
          swap = p[0]; p[0] = p[2]; p[2] = swap;
          set_CPU_bank4(p[0]);
          set_CPU_bank6(p[2]);
        }
        if(num_8k_ROM_banks)
        {
          if((data & 0x80) != (regs[6] & 0x80))
          {
            swap = chr4; chr4 = chr0; chr0 = swap;
            swap = chr5; chr5 = chr1; chr1 = swap;
            swap = chr6; chr6 = chr2; chr2 = swap;
            swap = chr7; chr7 = chr3; chr3 = swap;
            swap = c[4]; c[4] = c[0]; c[0] = swap;
            swap = c[5]; c[5] = c[1]; c[1] = swap;
            swap = c[6]; c[6] = c[2]; c[2] = swap;
            swap = c[7]; c[7] = c[3]; c[3] = swap;
            set_PPU_banks(c[0],c[1],c[2],c[3],c[4],c[5],c[6],c[7]);
          }
        }
        regs[6] = data;
      }
      break;

    case 0x8001:
      {
        switch(regs[6] & 0x07)
        {
          case 0x00:
            {
              chr0 = (data & 0xFE)+0;
              chr1 = (data & 0xFE)+1;
              MAP45_set_PPU_banks();
            }
            break;

          case 0x01:
            {
              chr2 = (data & 0xFE)+0;
              chr3 = (data & 0xFE)+1;
              MAP45_set_PPU_banks();
            }
            break;

          case 0x02:
            {
              chr4 = data;
              MAP45_set_PPU_banks();
            }
            break;

          case 0x03:
            {
              chr5 = data;
              MAP45_set_PPU_banks();
            }
            break;

          case 0x04:
            {
              chr6 = data;
              MAP45_set_PPU_banks();
            }
            break;

          case 0x05:
            {
              chr7 = data;
              MAP45_set_PPU_banks();
            }
            break;

          case 0x06:
            {
              if(regs[6] & 0x40)
              {
                prg2 = data & 0x3F;
                MAP45_set_CPU_bank6(data);
              }
              else
              {
                prg0 = data & 0x3F;
                MAP45_set_CPU_bank4(data);
              }
            }
            break;

          case 0x07:
            {
              prg1 = data & 0x3F;
              MAP45_set_CPU_bank5(data);
            }
            break;
        }
      }
      break;

    case 0xA000:
      {
        if(data & 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;

    case 0xC000:
      {
        irq_counter = data;
      }
      break;

    case 0xC001:
      {
        irq_latch = data;
      }
      break;

    case 0xE000:
      {
        irq_enabled = 0;
      }
      break;

    case 0xE001:
      {
        irq_enabled = 1;
      }
      break;
  }
}

void NES_mapper45::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(!(irq_counter--))
        {
          irq_counter = irq_latch;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}

void NES_mapper45::MAP45_set_CPU_bank4(uint8 data)
{
  data &= (regs[3] & 0x3F) ^ 0xFF;
  data &= 0x3F;
  data |= regs[1];
  set_CPU_bank4(data);
  p[0] = data;
}

void NES_mapper45::MAP45_set_CPU_bank5(uint8 data)
{
  data &= (regs[3] & 0x3F) ^ 0xFF;
  data &= 0x3F;
  data |= regs[1];
  set_CPU_bank5(data);
  p[1] = data;
}
void NES_mapper45::MAP45_set_CPU_bank6(uint8 data)
{
  data &= (regs[3] & 0x3F) ^ 0xFF;
  data &= 0x3F;
  data |= regs[1];
  set_CPU_bank6(data);
  p[2] = data;
}
void NES_mapper45::MAP45_set_CPU_bank7(uint8 data)
{
  data &= (regs[3] & 0x3F) ^ 0xFF;
  data &= 0x3F;
  data |= regs[1];
  set_CPU_bank7(data);
  p[3] = data;
}

void NES_mapper45::MAP45_set_PPU_banks()
{
  uint8 table[16] =
  {
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x01,0x03,0x07,0x0F,0x1F,0x3F,0x7F,0xFF
  };
  c[0] = chr0;
  c[1] = chr1;
  c[2] = chr2;
  c[3] = chr3;
  c[4] = chr4;
  c[5] = chr5;
  c[6] = chr6;
  c[7] = chr7;
  for(uint8 i = 0; i < 8; i++)
  {
    c[i] &= table[regs[2] & 0x0F];
    c[i] |= regs[0];
    c[i] += (regs[2] & 0x10) << 4;
  }
  if(regs[6] & 0x80)
  {
    set_PPU_banks(c[4],c[5],c[6],c[7],c[0],c[1],c[2],c[3]);
  }
  else
  {
    set_PPU_banks(c[0],c[1],c[2],c[3],c[4],c[5],c[6],c[7]);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 225
void NES_mapper225::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper225::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank = (addr & 0x0F80) >> 7;
  uint8 chr_bank = addr & 0x003F;

  set_PPU_bank0(chr_bank*8);
  set_PPU_bank1(chr_bank*8+1);
  set_PPU_bank2(chr_bank*8+2);
  set_PPU_bank3(chr_bank*8+3);
  set_PPU_bank4(chr_bank*8+4);
  set_PPU_bank5(chr_bank*8+5);
  set_PPU_bank6(chr_bank*8+6);
  set_PPU_bank7(chr_bank*8+7);

  if(addr & 0x2000)
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }
  else
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }

  if(addr & 0x1000)
  {
    // 16KBbank
    if(addr & 0x0040)
    {
      set_CPU_bank4(prg_bank*4+2);
      set_CPU_bank5(prg_bank*4+3);
      set_CPU_bank6(prg_bank*4+2);
      set_CPU_bank7(prg_bank*4+3);
    }
    else
    {
      set_CPU_bank4(prg_bank*4);
      set_CPU_bank5(prg_bank*4+1);
      set_CPU_bank6(prg_bank*4);
      set_CPU_bank7(prg_bank*4+1);
    }
  }
  else
  {
    set_CPU_bank4(prg_bank*4);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4+2);
    set_CPU_bank7(prg_bank*4+3);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 226
void NES_mapper226::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  regs[0] = 0;
  regs[1] = 0;
}

void NES_mapper226::MemoryWrite(uint32 addr, uint8 data)
{
  if(addr & 0x001)
  {
    regs[1] = data;
  }
  else
  {
    regs[0] = data;
  }

  if(regs[0] & 0x40)
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }
  else
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }

  uint8 prg_bank = ((regs[0] & 0x1E) >> 1) | ((regs[0] & 0x80) >> 3) | ((regs[1] & 0x01) << 5);

  if(regs[0] & 0x20)
  {
    if(regs[0] & 0x01)
    {
      set_CPU_bank4(prg_bank*4+2);
      set_CPU_bank5(prg_bank*4+3);
      set_CPU_bank6(prg_bank*4+2);
      set_CPU_bank7(prg_bank*4+3);
    }
    else
    {
      set_CPU_bank4(prg_bank*4);
      set_CPU_bank5(prg_bank*4+1);
      set_CPU_bank6(prg_bank*4);
      set_CPU_bank7(prg_bank*4+1);
    }
  }
  else
  {
    set_CPU_bank4(prg_bank*4);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4+2);
    set_CPU_bank7(prg_bank*4+3);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 227
void NES_mapper227::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,0,1);
}

void NES_mapper227::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank = ((addr & 0x0100) >> 4) | ((addr & 0x0078) >> 3);

  if(addr & 0x0001)
  {
    set_CPU_bank4(prg_bank*4);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4+2);
    set_CPU_bank7(prg_bank*4+3);
  }
  else
  {
    if(addr & 0x0004)
    {
      set_CPU_bank4(prg_bank*4+2);
      set_CPU_bank5(prg_bank*4+3);
      set_CPU_bank6(prg_bank*4+2);
      set_CPU_bank7(prg_bank*4+3);
    }
    else
    {
      set_CPU_bank4(prg_bank*4);
      set_CPU_bank5(prg_bank*4+1);
      set_CPU_bank6(prg_bank*4);
      set_CPU_bank7(prg_bank*4+1);
    }
  }
  if(!(addr & 0x0080))
  {
    if(addr & 0x0200)
    {
      set_CPU_bank6((prg_bank & 0x1C)*4+14);
      set_CPU_bank7((prg_bank & 0x1C)*4+15);
    }
    else
    {
      set_CPU_bank6((prg_bank & 0x1C)*4);
      set_CPU_bank7((prg_bank & 0x1C)*4+1);
    }
  }

  if(addr & 0x0002)
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }
  else
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 228
void NES_mapper228::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper228::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank = (addr & 0x0780) >> 7;
  uint8 chr_bank = ((addr & 0x000F) << 2) | (data & 0x03);

  switch((addr & 0x1800) >> 11)
  {
    case 1:
      {
        prg_bank |= 0x10;
      }
      break;

    case 3:
      {
        prg_bank |= 0x20;
      }
      break;
  }
  if(addr & 0x0020)
  {
    prg_bank = prg_bank << 1;
    if(addr & 0x0040)
    {
      prg_bank++;
    }
    set_CPU_bank4(prg_bank*4);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4);
    set_CPU_bank7(prg_bank*4+1);
  }
  else
  {
    set_CPU_bank4(prg_bank*4);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4+2);
    set_CPU_bank7(prg_bank*4+3);
  }

  set_PPU_bank0(chr_bank*8);
  set_PPU_bank1(chr_bank*8+1);
  set_PPU_bank2(chr_bank*8+2);
  set_PPU_bank3(chr_bank*8+3);
  set_PPU_bank4(chr_bank*8+4);
  set_PPU_bank5(chr_bank*8+5);
  set_PPU_bank6(chr_bank*8+6);
  set_PPU_bank7(chr_bank*8+7);

  if(addr & 0x2000)
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }
  else
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 229
void NES_mapper229::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper229::MemoryWrite(uint32 addr, uint8 data)
{
  addr &= 0x0FFF;

  if(addr & 0x0020)
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }
  else
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }

  if(addr & 0x001E)
  {
    uint8 prg_bank = (uint8)(addr & 0x001F);
    uint32 chr_bank = addr & 0x0FFF;
    set_CPU_bank4(prg_bank*2);
    set_CPU_bank5(prg_bank*2+1);
    set_CPU_bank6(prg_bank*2);
    set_CPU_bank7(prg_bank*2+1);
    set_PPU_bank0(chr_bank*8);
    set_PPU_bank1(chr_bank*8+1);
    set_PPU_bank2(chr_bank*8+2);
    set_PPU_bank3(chr_bank*8+3);
    set_PPU_bank4(chr_bank*8+4);
    set_PPU_bank5(chr_bank*8+5);
    set_PPU_bank6(chr_bank*8+6);
    set_PPU_bank7(chr_bank*8+7);
  }
  else
  {
    set_CPU_banks(0,1,2,3);
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 230
void NES_mapper230::Reset()
{
  // Contra - 22 Games switch
  if(rom_switch)
  {
    rom_switch = 0;
  }
  else
  {
    rom_switch = 1;
  }

  // set CPU bank pointers
  if(rom_switch)
  {
    set_CPU_banks(0,1,14,15);
  }
  else
  {
    set_CPU_banks(16,17,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  }
}

void NES_mapper230::MemoryWrite(uint32 addr, uint8 data)
{
  if(rom_switch)
  {
    set_CPU_bank4((data & 0x07)*2+0);
    set_CPU_bank5((data & 0x07)*2+1);
  }
  else
  {
    if(data & 0x40)
    {
      set_mirroring(NES_PPU::MIRROR_VERT);
    }
    else
    {
      set_mirroring(NES_PPU::MIRROR_HORIZ);
    }
    if(data & 0x20)
    {
      set_CPU_bank4((data & 0x1F)*2+16);
      set_CPU_bank5((data & 0x1F)*2+17);
      set_CPU_bank6((data & 0x1F)*2+16);
      set_CPU_bank7((data & 0x1F)*2+17);
    }
    else
    {
      set_CPU_bank4((data & 0x1E)*2+16);
      set_CPU_bank5((data & 0x1E)*2+17);
      set_CPU_bank6((data & 0x1E)*2+18);
      set_CPU_bank7((data & 0x1E)*2+19);
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 231
void NES_mapper231::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper231::MemoryWrite(uint32 addr, uint8 data)
{
  if(addr & 0x0020)
  {
    uint8 prg_bank = (uint8)(addr >> 1) & 0x0F;
    set_CPU_bank4(prg_bank*4);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4+2);
    set_CPU_bank7(prg_bank*4+3);
  }
  else
  {
    uint8 prg_bank = addr & 0x001E;
    set_CPU_bank4(prg_bank*2);
    set_CPU_bank5(prg_bank*2+1);
    set_CPU_bank6(prg_bank*2);
    set_CPU_bank7(prg_bank*2+1);
  }

  if(addr & 0x0080)
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }
  else
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 232
void NES_mapper232::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  prg_bank0 = 0x0C;
  prg_bank1 = 0x00;
}

void NES_mapper232::MemoryWrite(uint32 addr, uint8 data)
{
  if(addr == 0x9000)
  {
    prg_bank0 = (data & 0x18) >> 1;
  }
  else if(0xA000 <= addr && addr <= 0xFFFF)
  {
    prg_bank1 = data & 0x03;
  }
  uint8 prg0 = prg_bank0 | prg_bank1;
  uint8 prg1 = prg_bank0 | 0x03;

  set_CPU_bank4(prg0*2+0);
  set_CPU_bank5(prg0*2+1);
  set_CPU_bank6(prg1*2+0);
  set_CPU_bank7(prg1*2+1);
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 233
void NES_mapper233::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);
}

void NES_mapper233::MemoryWrite(uint32 addr, uint8 data)
{
  if(data & 0x20)
  {
    uint8 prg_bank = data & 0x1F;
    set_CPU_bank4(prg_bank*2);
    set_CPU_bank5(prg_bank*2+1);
    set_CPU_bank6(prg_bank*2);
    set_CPU_bank7(prg_bank*2+1);
  }
  else
  {
    uint8 prg_bank = (data & 0x1E) >> 1;
    set_CPU_bank4(prg_bank*4);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4+2);
    set_CPU_bank7(prg_bank*4+3);
  }

  if((data & 0xC0) == 0x00)
  {
    set_mirroring(0,0,0,1);
  }
  else if((data & 0xC0) == 0x40)
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }
  else if((data & 0xC0) == 0x80)
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }
  else
  {
    set_mirroring(1,1,1,1);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 234
void NES_mapper234::Reset()
{
  regs[0] = regs[1] = regs[2] = 0;

  set_CPU_banks(0,1,2,3);
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper234::MemoryWrite(uint32 addr, uint8 data)
{
  MemoryWrite2(addr, data);
}

void NES_mapper234::MemoryRead(uint32 addr)
{
  if((addr & 0xFF80) == 0xFF80)
  {
    MemoryWrite2(addr, parent_NES->cpu->GetByte(addr));
  }
}

void NES_mapper234::MemoryWrite2(uint32 addr, uint8 data)
{
  switch(addr & 0xFFF8)
  {
    case 0xFF80:
    case 0xFF88:
    case 0xFF90:
    case 0xFF98:
      {
        if(!regs[0])
        {
          regs[0] = data;
          if(data & 0x80)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
          if(data & 0x40)
          {
            set_CPU_bank4((data & 0x0E)*4+0);
            set_CPU_bank5((data & 0x0E)*4+1);
            set_CPU_bank6((data & 0x0E)*4+2);
            set_CPU_bank7((data & 0x0E)*4+3);
            set_PPU_bank0(((data & 0x0E)<<2)*8+0);
            set_PPU_bank1(((data & 0x0E)<<2)*8+1);
            set_PPU_bank2(((data & 0x0E)<<2)*8+2);
            set_PPU_bank3(((data & 0x0E)<<2)*8+3);
            set_PPU_bank4(((data & 0x0E)<<2)*8+4);
            set_PPU_bank5(((data & 0x0E)<<2)*8+5);
            set_PPU_bank6(((data & 0x0E)<<2)*8+6);
            set_PPU_bank7(((data & 0x0E)<<2)*8+7);
          }
          else
          {
            set_CPU_bank4((data & 0x0F)*4+0);
            set_CPU_bank5((data & 0x0F)*4+1);
            set_CPU_bank6((data & 0x0F)*4+2);
            set_CPU_bank7((data & 0x0F)*4+3);
            set_PPU_bank0(((data & 0x0F)<<2)*8+0);
            set_PPU_bank1(((data & 0x0F)<<2)*8+1);
            set_PPU_bank2(((data & 0x0F)<<2)*8+2);
            set_PPU_bank3(((data & 0x0F)<<2)*8+3);
            set_PPU_bank4(((data & 0x0F)<<2)*8+4);
            set_PPU_bank5(((data & 0x0F)<<2)*8+5);
            set_PPU_bank6(((data & 0x0F)<<2)*8+6);
            set_PPU_bank7(((data & 0x0F)<<2)*8+7);
          }
        }
      }
      break;

    case 0xFFC0:
    case 0xFFC8:
    case 0xFFD0:
    case 0xFFD8:
      {
        if(!regs[2])
        {
          regs[2] = data;
        }
      }
      break;

    case 0xFFE8:
    case 0xFFF0:
      {
        regs[1] = data;
        if (regs[0] & 0x40)
        {
          set_CPU_bank4(((regs[0] & 0x0E)|(regs[1] & 0x01))*4+0);
          set_CPU_bank5(((regs[0] & 0x0E)|(regs[1] & 0x01))*4+1);
          set_CPU_bank6(((regs[0] & 0x0E)|(regs[1] & 0x01))*4+2);
          set_CPU_bank7(((regs[0] & 0x0E)|(regs[1] & 0x01))*4+3);
          set_PPU_bank0((((regs[0] & 0x0E)<<2)|((regs[1] & 0x70)>>4))*8+0);
          set_PPU_bank1((((regs[0] & 0x0E)<<2)|((regs[1] & 0x70)>>4))*8+1);
          set_PPU_bank2((((regs[0] & 0x0E)<<2)|((regs[1] & 0x70)>>4))*8+2);
          set_PPU_bank3((((regs[0] & 0x0E)<<2)|((regs[1] & 0x70)>>4))*8+3);
          set_PPU_bank4((((regs[0] & 0x0E)<<2)|((regs[1] & 0x70)>>4))*8+4);
          set_PPU_bank5((((regs[0] & 0x0E)<<2)|((regs[1] & 0x70)>>4))*8+5);
          set_PPU_bank6((((regs[0] & 0x0E)<<2)|((regs[1] & 0x70)>>4))*8+6);
          set_PPU_bank7((((regs[0] & 0x0E)<<2)|((regs[1] & 0x70)>>4))*8+7);
        }
        else
        {
          set_CPU_bank4((regs[0] & 0x0F)*4+0);
          set_CPU_bank5((regs[0] & 0x0F)*4+1);
          set_CPU_bank6((regs[0] & 0x0F)*4+2);
          set_CPU_bank7((regs[0] & 0x0F)*4+3);
          set_PPU_bank0((((regs[0] & 0x0F)<<2)|((regs[1] & 0x30)>>4))*8+0);
          set_PPU_bank1((((regs[0] & 0x0F)<<2)|((regs[1] & 0x30)>>4))*8+1);
          set_PPU_bank2((((regs[0] & 0x0F)<<2)|((regs[1] & 0x30)>>4))*8+2);
          set_PPU_bank3((((regs[0] & 0x0F)<<2)|((regs[1] & 0x30)>>4))*8+3);
          set_PPU_bank4((((regs[0] & 0x0F)<<2)|((regs[1] & 0x30)>>4))*8+4);
          set_PPU_bank5((((regs[0] & 0x0F)<<2)|((regs[1] & 0x30)>>4))*8+5);
          set_PPU_bank6((((regs[0] & 0x0F)<<2)|((regs[1] & 0x30)>>4))*8+6);
          set_PPU_bank7((((regs[0] & 0x0F)<<2)|((regs[1] & 0x30)>>4))*8+7);
        }
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 235
void NES_mapper235::Reset()
{
  if(rom_switch)
  {
    rom_switch = 0;
  }
  else
  {
    rom_switch = 1;
  }

  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);
}

void NES_mapper235::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank;

  if(rom_switch)
  {
    prg_bank = ((addr & 0x0300) >> 3) | (addr & 0x001F);
  }
  else
  {
    if(num_8k_ROM_banks == 0x200)
    {
      prg_bank = ((addr & 0x0300) >> 3) | (addr & 0x001F);
    }
    else
    {
      prg_bank = ((addr & 0x0200) >> 4) | (addr & 0x001F);
    }
  }

  if(addr & 0x0400)
  {
    set_mirroring(0,0,0,0);
  }
  else if(addr & 0x2000)
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }
  else
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }

  if(addr & 0x0800)
  {
    if(addr & 0x1000)
    {
      // upper 16KB Banks
      set_CPU_bank4(prg_bank*4+2);
      set_CPU_bank5(prg_bank*4+3);
      set_CPU_bank6(prg_bank*4+2);
      set_CPU_bank7(prg_bank*4+3);
    }
    else
    {
      // lower 16KB Banks
      set_CPU_bank4(prg_bank*4+0);
      set_CPU_bank5(prg_bank*4+1);
      set_CPU_bank6(prg_bank*4+0);
      set_CPU_bank7(prg_bank*4+1);
    }
  }
  else
  {
    // 32KB Banks
    set_CPU_bank4(prg_bank*4+0);
    set_CPU_bank5(prg_bank*4+1);
    set_CPU_bank6(prg_bank*4+2);
    set_CPU_bank7(prg_bank*4+3);
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 255
void NES_mapper255::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);

  set_mirroring(NES_PPU::MIRROR_VERT);

  //regs[0] = regs[2] = 0x0F;
  //regs[1] = regs[3] = 0x00;
}

uint8 NES_mapper255::MemoryReadLow(uint32 addr)
{
  if(addr >= 0x5800)
  {
    return regs[addr & 0x0003] & 0x0F;
  }
  else
  {
    return (uint8)(addr >> 8);
  }
}

void NES_mapper255::MemoryWriteLow(uint32 addr, uint8 data)
{
  if(addr >= 0x5800)
  {
    regs[addr & 0x0003] = data & 0x0F;
  }
}

void NES_mapper255::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank = (uint8)((addr & 0x0F80) >> 7);
  uint8 chr_bank = (uint8)(addr & 0x003F);
  uint8 rom_bank = (uint8)((addr & 0x4000) >> 14);

  if(addr & 0x2000)
  {
    set_mirroring(NES_PPU::MIRROR_HORIZ);
  }
  else
  {
    set_mirroring(NES_PPU::MIRROR_VERT);
  }

  if(addr & 0x1000)
  {//16K PRG_ROM
    if(addr & 0x0040)
    {//upper half
      set_CPU_bank4(0x80*rom_bank+prg_bank*4+2);
      set_CPU_bank5(0x80*rom_bank+prg_bank*4+3);
      set_CPU_bank6(0x80*rom_bank+prg_bank*4+2);
      set_CPU_bank7(0x80*rom_bank+prg_bank*4+3);
    }
    else
    {//lower half
      set_CPU_bank4(0x80*rom_bank+prg_bank*4+0);
      set_CPU_bank5(0x80*rom_bank+prg_bank*4+1);
      set_CPU_bank6(0x80*rom_bank+prg_bank*4+0);
      set_CPU_bank7(0x80*rom_bank+prg_bank*4+1);
    }
  }
  else
  {//32K PRG_ROM
    set_CPU_bank4(0x80*rom_bank+prg_bank*4+0);
    set_CPU_bank5(0x80*rom_bank+prg_bank*4+1);
    set_CPU_bank6(0x80*rom_bank+prg_bank*4+2);
    set_CPU_bank7(0x80*rom_bank+prg_bank*4+3);
  }

  set_PPU_bank0(0x200*rom_bank+chr_bank*8+0);
  set_PPU_bank1(0x200*rom_bank+chr_bank*8+1);
  set_PPU_bank2(0x200*rom_bank+chr_bank*8+2);
  set_PPU_bank3(0x200*rom_bank+chr_bank*8+3);
  set_PPU_bank4(0x200*rom_bank+chr_bank*8+4);
  set_PPU_bank5(0x200*rom_bank+chr_bank*8+5);
  set_PPU_bank6(0x200*rom_bank+chr_bank*8+6);
  set_PPU_bank7(0x200*rom_bank+chr_bank*8+7);
}
/////////////////////////////////////////////////////////////////////

#endif
